###############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2018, 2019 XCG Consulting (http://www.xcg-consulting.fr/)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################
import logging
import os
import re
import sys
from base64 import b64encode
from typing import List

import requests

from odoo import _, api, exceptions, fields, models
from odoo.tools import config

from .util import get_secure_filename

if sys.platform == "win32":
    import pbs as sh
else:
    import sh

_logger = logging.getLogger(__name__)


class JobDefinition(models.Model):
    _name = "external_job.job_definition"
    _description = "External job definition"

    name = fields.Char(string="Name", size=128, required=True)

    command_line = fields.Char(
        string="Command Line", size=512, help="The program to call"
    )

    arguments_base = fields.Char(string="Arguments", size=2048)

    short_desc = fields.Char(
        string="Short Description", size=256, translate=True
    )

    long_desc = fields.Text(string="Long Description")

    output_filename_base = fields.Char(
        string="Output File Name",
        size=128,
        help=(
            "The desired file pattern . For example:"
            " %(job_definition.name)s-%(datetime.today)s.csv. If there is no"
            " value, no output file is created."
        ),
    )

    model_id = fields.Many2one(
        comodel_name="ir.model",
        string="Model",
        ondelete="cascade",
        help=(
            "The data model our job is expecting to process, used in the "
            "job runner wizard."
        ),
    )

    input_content = fields.Char(
        string="Content of the input file",
        help=(
            "A single argument, if the argument is a list, there will be"
            " one value per line. Without this argument, if no input file is "
            "given to run, no output file is created. If provided any input "
            "file will be ignored."
        ),
    )

    force_commit = fields.Boolean(
        string="Force Commit",
        help=(
            "Only needed when the external job need to see the job log "
            "in the database. Avoid as much as possible."
        ),
        default=False,
    )

    job_type = fields.Selection(
        selection=[("extrunner", "Extrunner"), ("cmd", "Command Line")],
        string="Job type",
        help="Choose the way to run the job.\n"
        " Extrunner: Call an external service to launch the job.\n"
        " Command line: Run the external process from odoo.",
        default="extrunner",
    )

    extrunner_server_id = fields.Many2one(
        "external_job.extrunner_server",
        string="Extrunner Server",
        ondelete="restrict",
    )

    job_ID = fields.Integer(
        string="Job ID", help="ID of the job to be launched by extrunner."
    )

    env_var_ids = fields.One2many(
        comodel_name="external_job.job_definition.env_var",
        inverse_name="job_def_id",
        string="Environment variables",
        help=(
            "Variables sent to extrunner. May contain dynamic arguments (see "
            "legend at the bottom of the job definition form)."
        ),
    )

    raise_error = fields.Boolean(
        string="Raise Error",
        help="Raise an exception when any error is found",
        default=False,
    )

    default_ids = fields.One2many(
        comodel_name="external_job.job_definition.default_value",
        string="Defaults",
        help="Dictionary to use as defaults (only implemented for context "
        "values)",
        inverse_name="job_def_id",
    )

    @api.multi
    def run_job(self, job_args: dict = None, in_file_name: str = None):
        """ This callback method is called by a user pressing a button
        in the form view. When the user presses the button we run the command
        and collect the data back for him.

        :param job_args: arguments for the job, added as a section of the dict
            of the string substitution
        :param in_file_name: name of the temporary input file to use (instead
            of generating one if needed)
        :return: a list of joblog_ids, to get a view change, use
            return_joblog_act_window, the order of the list is the same as
            for self.ids
        """
        joblog_ids = list()
        for job in self:
            if job.job_type == "extrunner":
                joblog_ids.extend(job._call_extrunner())
            if job.job_type == "cmd":
                joblog_ids.extend(job._call_cmd(job_args, in_file_name))

        return joblog_ids

    @staticmethod
    def return_joblog_act_window(joblog_ids: List[int]):
        """Return a view change from a list of job log ids"""
        if len(joblog_ids) == 1:
            return JobDefinition._return_joblog_act_window(
                joblog_ids, default_view="form"
            )
        return JobDefinition._return_joblog_act_window(
            joblog_ids, view_mode="tree"
        )

    @api.multi
    def _call_cmd(self, job_args: dict = None, in_file_name: str = None):

        cr, ctx = self.env.cr, self.env.context

        joblog_ids = []
        external_job_joblog = self.env["external_job.job_log"]
        for job in self:

            command = job.command_line
            job_log_vals = {"job_definition_id": job.id}

            if job.output_filename_base:
                out_file_name = get_secure_filename()
                job_log_vals["out_file_name"] = out_file_name

            if job.input_content:
                if in_file_name:
                    _logger.warning(
                        _("Ignoring provided filename %s in run job")
                        % in_file_name
                    )

                in_file_name = get_secure_filename()

            if in_file_name:
                job_log_vals["in_file_name"] = in_file_name

            joblog = external_job_joblog.create(job_log_vals)
            joblog_ids.append(joblog.id)

            arg_dict = ArgumentDict(
                job,
                joblog,
                ctx,
                job_args,
                cr,
                defaults=job.default_ids.to_dict(),
            )

            if job.input_content:
                # TODO handle exceptions
                with open(in_file_name, "w+") as f:
                    for value in arg_dict[job.input_content]:
                        f.write("%s\n" % value)

            # Make sure a commit of our transaction before giving the newly
            # created id to our external process or it won't be able to find it
            if job.force_commit:
                cr.commit()

            args = []
            if job.arguments_base:
                args.append(job.arguments_base % arg_dict)

            try:
                run = sh.Command(command)
            except sh.CommandNotFound:
                JobDefinition._handle_error(
                    joblog, _("Command %s not found") % command
                )
                return

            if job.arguments_base:
                out = run((job.arguments_base % arg_dict).split())
            else:
                out = run()

            # cleanup temporary files
            if in_file_name:
                os.remove(in_file_name)

            joblog.write(
                {
                    "end_date": fields.Datetime.now(),
                    "stdout": out,
                    "state": "done",
                }
            )

            if job.output_filename_base:
                with open(out_file_name, "rb") as outfile:
                    outfile_data = outfile.read()

                if not outfile_data:
                    outfile_data = ""

                outfile_data = b64encode(outfile_data)

                attachment = self.env["ir.attachment"]
                fname = job.output_filename_base % arg_dict
                attachment.create(
                    {
                        "name": fname,
                        "datas": outfile_data,
                        "datas_fname": fname,
                        "res_id": joblog.id,
                        "res_model": "external_job.job_log",
                        "type": "binary",
                    }
                )

                os.remove(out_file_name)

        return joblog_ids

    @staticmethod
    def _return_joblog_act_window(
        joblog_ids: List[int], view_mode="form", default_view="tree"
    ):
        """ return a dictionary that represent external job log action
        Note: default_view : if default view is form, specifies the
        record to load.
        """
        act = {
            "name": _("Job Log"),
            "type": "ir.actions.act_window",
            "res_model": "external_job.job_log",
            "view_mode": view_mode,
        }

        if len(joblog_ids) == 1 and default_view == "form":
            act["res_id"] = joblog_ids[0]
            return act

        act["domain"] = [("id", "in", joblog_ids)]
        return act

    @api.multi
    def _call_extrunner(self):
        """Call extrunner to launch an external job.
        """
        joblog_ids = []
        external_job_joblog = self.env["external_job.job_log"]

        for rec in self:
            job_log_vals = {"job_definition_id": rec.id}
            joblog = external_job_joblog.create(job_log_vals)
            joblog_ids.append(joblog.id)

            # Dynamic arguments: Similar to what we do in the "run_job" method;
            # not worth factoring out.
            arg_dict = ArgumentDict(
                rec,
                joblog,
                rec.env.context,
                None,  # "run" specifier not applicable here.
                rec.env.cr,
                defaults=rec.default_ids.to_dict(),
            )

            # login
            headers = {"Content-Type": "application/json"}
            server = rec.extrunner_server_id
            r = requests.post(
                "%s/auth/login" % server.base_url,
                json={
                    "login": server.user_login,
                    "password": server.user_password,
                },
                headers=headers,
            )
            if r.status_code != 200:
                JobDefinition._handle_error(
                    joblog,
                    _("status code: %s\n" "Failed to login to extrunner: %s")
                    % (r.status_code, r.text),
                )
                return

            headers["Authorization"] = "Bearer %s" % r.json().get(
                "id_token", "failed"
            )

            # Make sure a commit of our transaction before giving the newly
            # created id to our external process or it won't be able to find it
            if rec.force_commit:
                rec.env.cr.commit()

            # Include our env var key/value pairs into the JSON data sent here.
            # Values support dynamic formatting via "ArgumentDict".
            data = {
                "env": {
                    var.key: var.value % arg_dict for var in rec.env_var_ids
                }
            }
            data["env"].update({"dbname": rec.env.cr.dbname})

            args = rec.env.context.get("args")
            if args:
                for arg in re.split("--", args):
                    res = arg.split()
                    if res:
                        data["env"].update({res[0]: res[1]})

            r = requests.post(
                "%s/run/%d" % (server.base_url, rec.job_ID),
                headers=headers,
                json=data,
            )

            if r.status_code != 200:
                if r.status_code == 400:
                    # Bad request
                    JobDefinition._handle_error(
                        joblog,
                        _("Bad status code: %s\n" "Bad request: %s")
                        % (r.status_code, r.text),
                    )
                if r.status_code == 401:
                    # Unauthorized
                    JobDefinition._handle_error(
                        joblog,
                        _("Bad status code: %s\n" "Unauthorized")
                        % r.status_code,
                    )
                if r.status_code == 520:
                    # Error server
                    JobDefinition._handle_error(
                        joblog,
                        _("Bad status code: %s\n" "Server internal error: %s")
                        % (r.status_code, r.text),
                    )
                continue

            values = {"end_date": fields.Datetime.now(), "state": "done"}
            # TODO maybe test r.content
            if rec.output_filename_base:
                fname = rec.output_filename_base % arg_dict
                attachment = rec.env["ir.attachment"].create(
                    {
                        "name": fname,
                        "datas": b64encode(r.content),
                        "datas_fname": fname,
                        "res_id": joblog.id,
                        "res_model": "external_job.job_log",
                        "type": "binary",
                    }
                )
                values["attachment_id"] = attachment.id

            joblog.write(values)

        return joblog_ids

    @staticmethod
    def _handle_error(joblog, msg: str):
        joblog.write(
            {
                "end_date": fields.Datetime.now(),
                "state": "exception",
                "stderr": msg,
            }
        )
        if joblog.job_definition_id.raise_error:
            joblog.env.cr.commit()
            raise exceptions.UserError(_("Error"), msg)


class ArgumentDict(object):
    """Used to create arguments of a job.
    """

    def __init__(self, job_definition, joblog, context, run, cr, defaults={}):
        self.job_definition = job_definition
        self.joblog = joblog
        self.context = context
        self.run = run
        self.cr = cr
        self.defaults = defaults

    # incorrect but does not seem to be used
    def __len__(self):
        return 3 + len(self.context) + len(self.run)

    def __getitem__(self, key):
        parts = key.split(".")
        if parts[0] == "job_log":
            # dangerous?
            return getattr(self.joblog, parts[1])
        if parts[0] == "job_definition":
            # dangerous?
            return getattr(self.job_definition, parts[1])
        if parts[0] == "context":
            if self.context:
                if parts[1] in self.context:
                    return self.context[parts[1]]
                else:
                    return self.defaults[key]
        if parts[0] == "run":
            return self.run[parts[1]]
        if parts[0] == "config":
            if parts[1] == "options":
                if parts[2] == "db_port":
                    return (
                        config.options[parts[2]]
                        if config.options[parts[2]]
                        else 5432
                    )
                else:
                    return config.options[parts[2]]
            else:
                return config.misc.get(parts[1], {}).get(parts[2])
        if parts[0] == "cr":
            if parts[1] == "dbname":
                return self.cr.dbname
        if parts[0] == "datetime":
            if parts[1] == "now":
                return fields.Datetime.now()
            elif parts[1] == "today":
                return fields.Date.today()

    # does not seem to be used
    def __contains__(self, item):
        if item in [
            "job_log.id",
            "job_log.in_file_name",
            "job_log.out_file_name",
        ]:
            return True
        parts = item.split(".")
        if parts[0] == "job_definition":
            # TODO do real test
            return True
        if parts[0] == "context":
            return self.context and (
                parts[1] in self.context or item in self.defaults
            )
        if parts[0] == "run":
            return self.run and parts[1] in self.run
        if parts[0] == "config":
            if parts[1] == "options":
                return parts[2] in config.options
            else:
                return parts[2] in config.misc.get(parts[1], {})
        if parts[0] == "cr":
            if parts[1] == "dbname":
                return True
        return False

    # XXX __iter__ not used apparently
