###############################################################################
#
#    External Job, for Odoo
#    Copyright (C) 2018 XCG Consulting (http://www.xcg-consulting.fr/)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

import base64

import mock  # Odoo req
import requests  # Odoo req

from odoo.tests import TransactionCase


class Test(TransactionCase):
    """Test external job definition.

    These tests assume an Odoo database with demo data
    """

    def setUp(self):
        super(Test, self).setUp()
        self.user_erp_manager = self.env.ref("external_job.user_erp_manager")

    def test_create_job_definition(self):
        """Create a new job definition with an erp manager.
        Assert the job definition is created (id>0).
        """

        new_job_definition = (
            self.env["external_job.job_definition"]
            .sudo(self.user_erp_manager)
            .create({"name": "test job definition", "command_line": "ls"})
        )

        # Ensure oour job def has been created successfully
        self.assertTrue(new_job_definition)

    def test_run_job_echo_1(self):
        """Run an echo 1 job and assert that the result id is in the output.
        """

        external_job_def = self.env["external_job.job_definition"]
        external_job_log = self.env["external_job.job_log"]

        echo_job = self.ref("external_job.job_definition_echo_1")
        job_def = external_job_def.browse(echo_job)
        job_log_ids = job_def.run_job()

        job_log = external_job_log.browse(job_log_ids)
        self.assertEqual("%d\n" % job_log_ids[0], job_log.stdout)

    def test_run_job_echo_2(self):
        """Run an echo 2 job and assert that the output is correct.
        """

        external_job_def = self.env["external_job.job_definition"]
        external_job_log = self.env["external_job.job_log"]

        echo_job = self.ref("external_job.job_definition_echo_2")
        job_def = external_job_def.browse(echo_job)
        job_log_ids = job_def.run_job()

        job_log = external_job_log.browse(job_log_ids)
        self.assertEqual("echo 2\n", job_log.stdout)

    def test_run_job_cat(self):
        """Run a cat job and assert the output is the content of the file
        """

        external_job_def = self.env["external_job.job_definition"]
        external_job_log = self.env["external_job.job_log"]

        cat_job = self.ref("external_job.job_definition_cat")
        job_def = external_job_def.browse(cat_job)

        job_log_ids = job_def.with_context(values=[1, 2, 3]).run_job()

        job_log = external_job_log.browse(job_log_ids)
        self.assertEqual("1\n2\n3\n", job_log.stdout)

    @mock.patch("requests.post")
    def test_run_job_extrunner(self, requests_post_mock):
        """Run an extrunner job.
        """

        # Mock the HTTP POST request.
        requests_post_mock.return_value = mock.Mock(
            content=b"ATTACHMENT-CONTENTS", status_code=requests.codes.ok
        )

        server = self.env["external_job.extrunner_server"].create(
            {"base_url": "http://BASE-URL-1"}
        )
        # Create an extrunner job def.
        job_def = self.env["external_job.job_definition"].create(
            {
                "extrunner_server_id": server.id,
                "env_var_ids": [
                    (0, 0, {"key": "STATIC-1", "value": "1"}),
                    (
                        0,
                        0,
                        {"key": "DYN-1", "value": "%(job_definition.name)s x"},
                    ),
                ],
                "job_type": "extrunner",
                "name": "DEF-1",
                "output_filename_base": "test.csv",
            }
        )
        self.assertTrue(job_def)

        def getJobs():
            """Get jobs created out of our extrunner job def."""
            return self.env["external_job.job_log"].search(
                [("job_definition_id", "=", job_def.id)]
            )

        # Run it!
        self.assertEqual(len(getJobs()), 0)  # no jobs yet.
        job_def.sudo(self.ref("base.user_demo")).run_job()
        job = getJobs()
        self.assertEqual(len(job), 1)  # 1 new job.

        # Our POST mock should have been called twice (auth & run).
        requests_post_mock.assert_any_call(
            "http://BASE-URL-1/auth/login", json=mock.ANY, headers=mock.ANY
        )
        requests_post_mock.assert_any_call(
            "http://BASE-URL-1/run/0", json=mock.ANY, headers=mock.ANY
        )

        # Check key/value env vars sent in the last call. See job_def creation
        # above for env vars we test here.
        last_call_data = requests_post_mock.call_args[-1]["json"]
        self.assertEqual(
            last_call_data,
            {
                "env": {
                    "dbname": self.env.cr.dbname,
                    "STATIC-1": "1",
                    "DYN-1": "DEF-1 x",  # jobdef name & " x"
                }
            },
        )

        # Check the new job.
        self.assertTrue(job.end_date)
        self.assertEqual(job.state, "done")

        # Ensure the new job has an attachment with contents from the mock.
        att = self.env["ir.attachment"].search(
            [("res_id", "=", job.id), ("res_model", "=", job._name)]
        )
        self.assertEqual(len(att), 1)
        self.assertEqual(att.datas, base64.b64encode(b"ATTACHMENT-CONTENTS"))
        self.assertTrue(job.attachment_id)
